;
;  A few standard functions
;

; the usual c[ad]*r series, generated with a script :)
(define (caar x)
  (car (car x)))
(define (cadr x)
  (car (cdr x)))
(define (cdar x)
  (cdr (car x)))
(define (cddr x)
  (cdr (cdr x)))
(define (caaar x)
  (car (car (car x))))
(define (caadr x)
  (car (car (cdr x))))
(define (cadar x)
  (car (cdr (car x))))
(define (caddr x)
  (car (cdr (cdr x))))
(define (cdaar x)
  (cdr (car (car x))))
(define (cdadr x)
  (cdr (car (cdr x))))
(define (cddar x)
  (cdr (cdr (car x))))
(define (cdddr x)
  (cdr (cdr (cdr x))))
(define (caaaar x)
  (car (car (car (car x)))))
(define (caaadr x)
  (car (car (car (cdr x)))))
(define (caadar x)
  (car (car (cdr (car x)))))
(define (caaddr x)
  (car (car (cdr (cdr x)))))
(define (cadaar x)
  (car (cdr (car (car x)))))
(define (cadadr x)
  (car (cdr (car (cdr x)))))
(define (caddar x)
  (car (cdr (cdr (car x)))))
(define (cadddr x)
  (car (cdr (cdr (cdr x)))))
(define (cdaaar x)
  (cdr (car (car (car x)))))
(define (cdaadr x)
  (cdr (car (car (cdr x)))))
(define (cdadar x)
  (cdr (car (cdr (car x)))))
(define (cdaddr x)
  (cdr (car (cdr (cdr x)))))
(define (cddaar x)
  (cdr (cdr (car (car x)))))
(define (cddadr x)
  (cdr (cdr (car (cdr x)))))
(define (cdddar x)
  (cdr (cdr (cdr (car x)))))
(define (cddddr x)
  (cdr (cdr (cdr (cdr x)))))

; versions of car/cdr which do not fail when handed something other than a cons
(define (safe-car c)
  (if (consp c)
      (car c)
    nil))

(define (safe-cdr c)
  (if (consp c)
      (cdr c)
    nil))

; map FUN over elements of LIST, returns list of results.
(define (mapcar fun list)
  (if list
      (cons (fun (car list)) (mapcar fun (cdr list))))
  )

; append two lists by copying the first list.
(define (append a b)
  (if a
      (cons (car a) (append (cdr a) b))
    b))

; get nth element of a list, indexing starts with 0.
(define (nth n list)
  (if (> n 0)
      (nth (- n 1) (safe-cdr list))
    (safe-car list)))

; set nth element of a list.
(define (set-nth n list value)
  (if (consp list)
      (if (> n 0)
          (set-nth (- n 1) (safe-cdr list) value)
        (setcar list value))
    (throw 'type-error "argument is not a cons")
    )
  )

(define-setf nth set-nth)

; get length of list.
(define (length list)
  (if list
      (+ 1 (length (cdr list)))
    0))

(define (curry-aux list)
  (if list
      (cons (cons 'quote
                  (cons (eval (car list)) nil))
            (curry-aux (cdr list)))))

; simple currying.
; (curry '(fun arg1 arg2) 3) returns a function that takes three
; more arguments arg3, arg4, arg5 and then returns (fun arg1 arg2
; arg3 arg4 arg5).
(define (curry expr n)
  (let* ((fun (car expr))
         (args (cdr expr))
         (anum (length args))
         (total (+ anum n))
         newargs
         (i 0))
    (while (< i n)
      (set 'newargs (cons (gensym) newargs)
           'i (+ 1 i))
      )
    (eval
     (cons 'lambda
           (cons newargs
                 (cons (append (curry-aux expr) newargs)
                       nil))))
    )
  )

; non-nil if ELT exists in LIST
(define (memq elt list)
  (while (and list
              (not (eq elt (car list))))
    (set 'list (cdr list)))
  list)

; apply FUN to every value between MIN and MAX. Replacement for simple
; for loops. Returns concatenation of results.
(define (map-range min max fun)
  (if (<= min max)
      (cons (fun min)
            (map-range (+ 1 min) max fun))
    nil))

; Association lists: modeled after emacs' functions
(define (assoc key alist)
  (catch :found
    (while alist
      (if (eq (caar alist) key)
          (throw :found (cdar alist)))
      (setq alist (cdr alist)))))

(define (rassoc value alist)
  (catch :found
    (while alist
      (if (eq (cdar alist) value)
          (throw :found (caar alist)))
      (setq alist (cdr alist)))))

(define (remassoc key alist)
  (if (not alist)
      nil
    (if (eq (caar alist) key)
        (cdr alist)
      (setcdr alist (remassoc key (cdr alist)))   ; inefficient, but simple
      alist)))

(define (copy-alist alist)
  (if alist
      (cons (cons (caar alist) (cdar alist)) (copy-alist alist))))

